<?php

namespace FiloBlu\Flow\Model\From;

use Exception;
use FiloBlu\Flow\Helper\Data;
use FiloBlu\Flow\Helper\EmailNotification;
use FiloBlu\Flow\Helper\LoggerProvider;
use Magento\Catalog\Model\ProductFactory;
use Magento\CatalogInventory\Api\StockConfigurationInterface;
use Magento\CatalogInventory\Api\StockRegistryInterface;
use Magento\CatalogInventory\Model\Stock as CatalogInventoryStock;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\DB\Adapter\AdapterInterface;
use Magento\Framework\Exception\MailException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Model\Context;
use Magento\Framework\Registry;
use Monolog\Logger;
use Zend_Db;

use Magento\InventoryApi\Api\SourceItemsSaveInterface;
use Magento\InventoryApi\Api\Data\SourceItemInterfaceFactory;

/**
 * Class Multi Invetory (tested only on 2.3.2)
 * @package FiloBlu\Flow\Model\From
 * https://devdocs.magento.com/guides/v2.3/inventory/
 */
class Multistock extends AbstractFrom
{

    /**
     * @var Logger
     */
    protected $_logger;

    /**
     * @var Data
     */
    protected $helperFlow;

    /**
     * @var ProductFactory
     */
    protected $productFactory;

    /**
     * @var ProductMetadataInterface
     */
    protected $_productMetadata;

    /**
     * @var EmailNotification
     */
    protected $emailNotification;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var
     */
    protected $connection;

    /**
     * @var StockConfigurationInterface
     */
    protected $stockConfiguration;

    protected $_sourceItemsSaveInterface;
    protected $_sourceItemFactory;

    // Various names for the default source
    protected $defaultMagentoSourceNames = array('snatt','default','magento','ve');

    public function __construct(
        Context $context,
        Registry $registry,
        Data $helper_flow,
        LoggerProvider $loggerProvider,
        ProductFactory $productFactory,
        StockConfigurationInterface $stockConfiguration,
        ProductMetadataInterface $productMetadata,
        EmailNotification $emailNotification,
        ScopeConfigInterface $scopeConfig,
        SourceItemsSaveInterface $sourceItemsSaveInterface,
        SourceItemInterfaceFactory $sourceItemFactory,
        array $data = []
    ) {
        $this->helperFlow = $helper_flow;
        $this->_logger = $loggerProvider->getLogger();
        $this->productFactory = $productFactory;
        $this->_productMetadata = $productMetadata;
        $this->emailNotification = $emailNotification;
        $this->scopeConfig = $scopeConfig;
        $this->stockConfiguration = $stockConfiguration;
        $this->_sourceItemsSaveInterface = $sourceItemsSaveInterface;
        $this->_sourceItemFactory = $sourceItemFactory;
        parent::__construct($context, $registry, null, null, $data);
    }

    /**
     * @return void
     */
    public function _construct()
    {
        $this->_init(\FiloBlu\Flow\Model\ResourceModel\From\Stock::class);
    }

    /**
     * @return $this|bool|mixed
     * @throws Exception
     */
    public function process()
    {
        $sku = $this->getSku();
        $productId = $this->productFactory->create()->getIdBySku($sku);
        $magazzino = $this->getData('magazzino');
        $qty = intval($this->getData('fisico_disponibile'));

        if (!$productId) {
            return $this->error(false, "Product not found! product_code: '{$sku}'");
        }

        if (!$this->getData('magazzino')) {
            return $this->error(false, "Magazzino code not found! product_code: '{$sku}'");
        }

        try {
            //create the sourceItem using the factory
            $sourceItem = $this->_sourceItemFactory->create();
            $sourceItem->setSourceCode($magazzino);
            $sourceItem->setSku($sku);
            $sourceItem->setQuantity($qty);
            if($qty > 0){
                $sourceItem->setStatus(1);
            } else {
                $sourceItem->setStatus(0);
            }

            //pass the sourceItem as an array element, you can add more source items in the same call as further array elements
            $this->_sourceItemsSaveInterface->execute([$sourceItem]);

            $this->_logger->info("Process STOCK - qty {$qty} ID: {$productId} SKU: {$sku}");
        } catch (Exception $e) {
            $this->_logger->info('Errore nel salvataggio dello Stock. ' . $e->getMessage());
        }

        return $this;
    }

    /**
     * @return false|AdapterInterface
     */
    protected function getConnection()
    {
        if ($this->connection) {
            return $this->connection;
        } else {
            return ($this->connection = $this->getResource()->getConnection());
        }
    }

    /**
     * @param $query
     * @return bool
     */
    protected function postprocessFileRows($query)
    {
        try {
            $this->updateConfigurablesStockStatus();
        } catch (Exception $e) {
            $this->_logger->info('Errore nel salvataggio del Post Stock Item Fix Tuning. ' . $e->getMessage());
            return false;
        }

        return true;
    }

    /**
     * 1] $this->stockConfiguration->getBackorders($this->getStoreId()) = 0
     *      backorders = 0 && use_config_backorders = 1
     *
     * 2]  backorders = 0 && use_config_backorders = 0
     */
    protected function updateConfigurablesStockStatus()
    {
        $connection = $this->getConnection();
        $entityField = $this->getEntityField();

        $sql = $connection->select()
            ->from(['isi' => 'inventory_source_item'], [])
            ->joinLeft(['cpe' => 'catalog_product_entity'], 'cpe.sku = isi.sku', [])
            ->joinLeft(['cpsl' => 'catalog_product_super_link'], 'cpe.entity_id = cpsl.product_id', ['cpsl.parent_id'])
            ->where('cpsl.parent_id is not null')
            ->group('cpsl.parent_id');
        // Non sono sicuro del group by parent id

        $outOfStockSql = clone $sql;
        $inStockSql = clone $sql;

        $outOfStockSql->having('SUM(isi.quantity) <= 0');
        $inStockSql->having('SUM(isi.quantity) > 0');

        $writer = new \Zend\Log\Writer\Stream(BP . '/var/log/source_analysis.log');
        $logger = new \Zend\Log\Logger();
        $logger->addWriter($writer);

        /* 1] Update Out Of Stock configurables */
        $products = $connection->select()
            ->from('catalog_product_entity', ['entity_id'])
            ->where('type_id = ?', ['configurable'])
            ->where("{$entityField} IN (?)", [$outOfStockSql])
            ->query()->fetchAll(Zend_Db::FETCH_NUM);

        if (count($products)){
            $configurableIds = [];
            foreach ($products as $row) {
                $configurableIds[] = $row[0];
            }
            $ids = implode(',', $configurableIds);
            $connection->query("UPDATE cataloginventory_stock_item SET is_in_stock = 0 WHERE `product_id` IN ({$ids})");
        }


        /* 2] Update In Stock configurables */
        $products = $products = $connection->select()
            ->from('catalog_product_entity', ['entity_id'])
            ->where('type_id = ?', ['configurable'])
            ->where("{$entityField} IN (?)", [$inStockSql])
            ->query()->fetchAll(Zend_Db::FETCH_NUM);

        if (count($products)) {
            $configurableIds = [];
            foreach ($products as $row) {
                $configurableIds[] = $row[0];
            }
            $ids = implode(',', $configurableIds);
            $connection->query("UPDATE cataloginventory_stock_item SET is_in_stock = 1 WHERE `product_id` IN ({$ids})");
        }
    }

    /**
     * @return string
     */
    protected function getEntityField()
    {
        $edition = $this->_productMetadata->getEdition();
        return ($edition === 'Enterprise' || $edition === 'B2B') ? 'row_id' : 'entity_id';
    }

    public function sendErrorNotifications($file){
        // TODO
        return true;
    }
}
